#include "console_input.h"

#include "ch.h"
#include "hal.h"
#include <string.h>

//#define USE_JTAG_SWD

#include "ump_frame.h"
#include "lpc1114_ump.h"
static struct UmpFrame eventFrame;

static Mailbox *g_inputEventMailbox = 0;

#define KEY_MESSAGE(key,pressed) ((4<<24) | ((pressed)?(1<<23):0) | (key))
#define FADERKEY_MESSAGE(faderid,key,pressed) ((2<<24) | ((faderid)<<16) | ((pressed)<<15) |  ((unsigned short)(key)))
#define FADER_MESSAGE(faderid,value) ((1<<24) | ((faderid)<<16) | ((unsigned short)(value)))

int abs (int _value)
{
    return (_value < 0) ? (-_value) : _value;
}

uint16_t median(const uint16_t * _values, const int N)
{
    int i;
    int unsorted = 1;
    uint16_t values[N];

    for (i = 0; i < N; ++i)
    {
	values[i] = _values[i];
    }

    while(unsorted)
    {
	unsorted = 0;
	for (i=0; i < N-1; ++i)
	{
	    if (values[i] > values[i+1])
	    {
		const int t = values[i];
		values[i] = values[i+1];
		values[i+1] = t;
		unsorted = 1;
	    }
	}
    }
    return values[(N-1)/2];
}


/*
 *
 * Faders:
 *  1:PIO0_11 - ADC0
 *  2:PIO1_0  - ADC1
 *  3:PIO1_1  - ADC2
 *  4:PIO1_2  - ADC3
 *  5:PIO1_4  - ADC5
 */
/*
 * Keyboard:
 *  COL1: PIO0_3
 *  COL2: PIO0_8
 *  COL3: PIO0_9
 *  COL4: PIO0_2
 *  COL5: PIO0_0 / RESET
 *
 *  ROW1: PIO0_1
 *  ROW2: PIO0_6
 *  ROW3: PIO1_5
 *
 *  LED_ROW1: PIO0_7
 *  LED_ROW2: PIO1_9
 *  LED_ROW3: PIO1_8
 *  LED_ROW4: PIO1_3 / SWDIO
 *  LED_ROW5: PIO0_10 / SWCLK
 */

enum {
    TIMER_ENABLE = (1 << 0),
    TIMER_RESET = (1 << 1)
};


static void Chip_TIMER_Reset(LPC_TMR_TypeDef *pTMR)
{
    uint32_t reg;

    /* Disable timer, set terminal count to non-0 */
    reg = pTMR->TCR;
    pTMR->TCR = 0;
    pTMR->TC = 1;

    /* Reset timer counter */
    pTMR->TCR = TIMER_RESET;

    /* Wait for terminal count to clear */
    while (pTMR->TC != 0) {}

    /* Restore timer state */
    pTMR->TCR = reg;
}

void Chip_TIMER_Disable(LPC_TMR_TypeDef *pTMR)
{
    pTMR->TCR &= ~TIMER_ENABLE;
}

void Chip_TIMER_Enable(LPC_TMR_TypeDef *pTMR)
{
    pTMR->TCR |= TIMER_ENABLE;
}

static void Chip_TIMER_SetMatch(LPC_TMR_TypeDef *pTMR, int8_t matchnum, uint32_t matchval)
{
    ((uint32_t*)&pTMR->MR0)[matchnum] = matchval;
    return;

    switch (matchnum)
    {
    case 0:
        pTMR->MR0 = matchval;
	break;

    case 1:
        pTMR->MR1 = matchval;
	break;

    case 2:
        pTMR->MR2 = matchval;
	break;

    case 3:
        pTMR->MR3 = matchval;
	break;
    }
}

int Chip_TIMER_MatchPending(LPC_TMR_TypeDef *pTMR, const uint8_t _matchnum)
{
    return (pTMR->IR & (1<<_matchnum)) != 0;
}

static inline void Chip_TIMER_ClearMatch(LPC_TMR_TypeDef *pTMR, uint32_t _channel)
{
    pTMR->IR = 1 << _channel;
}

static inline void Chip_TIMER_PWMWrite(LPC_TMR_TypeDef *pTMR, uint32_t pwmval)
{
        pTMR->PWMC = pwmval;
}

static inline uint32_t Chip_TIMER_PWMRead(LPC_TMR_TypeDef *pTMR)
{
        return pTMR->PWMC;
}



enum
{
	KeyboardRowCount = 3,
	KeyboardLedRowCount = 5,
	KeyboardColumnCount = 5,
	KeyboardCellCount = KeyboardRowCount * KeyboardColumnCount,
	KeyboardLedStateCount = 4,
	FaderCount = KeyboardColumnCount,
	MedianFilterSize = 14
};

enum
{
    KeyboardLed_TimerResolution = 1000/*µs*/,
    KeyboardLed_TimerInactiveValue = 10000 /*µs*/
};

const char ColumnSelectedLevels[5]   = {1,1,1,1,1}; // N-Channel MOSFETs on all columns.
const char ColumnDeselectedLevels[5] = {0,0,0,0,0};

static volatile unsigned short ledValues[KeyboardLedStateCount][KeyboardColumnCount][KeyboardLedRowCount];
static unsigned long currentKeyState = 0;
static unsigned char currentKeyCollumn = 0;
static unsigned char currentBlinkState = 0;
static unsigned char currentBlinkTimeout = 0;

static const int columnPorts[KeyboardColumnCount] = {0, 0, 0, 0, 0};
static const int columnPins[KeyboardColumnCount]  = {3, 8, 9, 2, 0};

static const int ledRowPorts[KeyboardLedRowCount] = {0, 1, 1, 1,  0};
static const int ledRowPins[KeyboardLedRowCount]  = {7, 9, 8, 3, 10};
//static const int ledRowPorts[KeyboardLedRowCount] = {1, 1,  0, 0, 1};
//static const int ledRowPins[KeyboardLedRowCount]  = {8, 3, 10, 7, 9};

static const int kbdRowPorts[KeyboardRowCount] = {0, 0, 1};
static const int kbdRowPins[KeyboardRowCount]  = {1, 6, 5};

static void InitializeAdc(void)
{
    // enable ADC
    LPC_SYSCON->SYSAHBCLKCTRL |= (1 << 13);
    LPC_SYSCON->PDRUNCFG &= ~(1<<4);
    LPC_IOCON->R_PIO0_11 = 0x40 | (0<<3) | 2; // AD0: ADC pull-down
    LPC_IOCON->R_PIO1_0  = 0x40 | (0<<3) | 2; // AD1: ADC pull-down
    LPC_IOCON->R_PIO1_1  = 0x40 | (0<<3) | 2; // AD2: ADC pull-down
    LPC_IOCON->R_PIO1_2  = 0x40 | (0<<3) | 2; // AD3: ADC pull-down (SWDIO)
    LPC_IOCON->PIO1_4    = 0x40 | (0<<3) | 1; // AD3: ADC pull-down
    LPC_ADC->INTEN = 0;
    
    const unsigned char channelmask = (1<<0) | (1<<1) | (1<<2) | (1<<3) | (1<<5); // AD0
    const unsigned char adcClkDiv = 255; // 16; // clock divider
    LPC_ADC->CR = channelmask | (adcClkDiv<<8) | (1<<16); // BURST-Mode

    /*
     * in burst mode the adc are scanned by hardware and there is no need to start a conversion.
     * If START is set, the burst converstion will not start
     *-- LPC_ADC->CR |= 1<<24; // adc_startConversion();
     */
}

static int FaderGetAdcValue(const int _column, uint16_t * _value)
{
    const uint32_t v = LPC_ADC->DR[(_column<4) ? _column : 5];
    const int done = (v & (1<< 31)) ? 1 : 0;
    if (done)
    {
	*_value =  (v >> 6) & 0x3FF;
    }
    return done;
}


static void InitializeKeyboardPinmux(void)
{
    //-- setup column selector (outputs)
    palSetGroupMode(GPIO0, PAL_PORT_BIT(3) | PAL_PORT_BIT(8) | PAL_PORT_BIT(9) | PAL_PORT_BIT(2) | PAL_PORT_BIT(0),
		    0, PAL_MODE_OUTPUT_PUSHPULL);
    LPC_IOCON->PIO0_3  = 0xC0 | (0<<3) | 0; // COL1 - Digital function GPIO
    LPC_IOCON->PIO0_8  = 0xC0 | (0<<3) | 0; // COL2 - Digital function GPIO
    LPC_IOCON->PIO0_9  = 0xC0 | (0<<3) | 0; // COL3 - Digital function GPIO
    LPC_IOCON->PIO0_2  = 0xC0 | (0<<3) | 0; // COL4 - Digital function GPIO
    LPC_IOCON->RESET_PIO0_0  = 0xC0 | (0<<3) | 1; // COL5 - Digital function GPIO (RESET)

    //-- select rows (inputs)
    palSetGroupMode(GPIO0, PAL_PORT_BIT(1) | PAL_PORT_BIT(6),
		    0, PAL_MODE_INPUT);
    palSetGroupMode(GPIO1, PAL_PORT_BIT(5),
		    0, PAL_MODE_INPUT);
    LPC_IOCON->PIO0_1        = 0xC0 | (2<<3) | 0; // ROW1 - Digital function GPIO pull-up
    LPC_IOCON->PIO0_6        = 0xC0 | (2<<3) | 0; // ROW2 - Digital function GPIO pull-up
    LPC_IOCON->PIO1_5        = 0xC0 | (2<<3) | 0; // ROW3 - Digital function GPIO pull-up

    //-- LED-Rows
    palSetGroupMode(GPIO0, PAL_PORT_BIT(7) | PAL_PORT_BIT(10),
		    0, PAL_MODE_OUTPUT_PUSHPULL);
    palSetGroupMode(GPIO1, PAL_PORT_BIT(8) | PAL_PORT_BIT(9) | PAL_PORT_BIT(3),
		    0, PAL_MODE_OUTPUT_PUSHPULL);
    LPC_IOCON->PIO0_7        = 0xC0 | (0<<3) | 0; // LED_ROW1: PIO0_7
    LPC_IOCON->PIO1_9        = 0xC0 | (0<<3) | 0; // LED_ROW2: PIO1_9
    LPC_IOCON->PIO1_8        = 0xC0 | (0<<3) | 0; // LED_ROW3: PIO1_8
#ifndef USE_JTAG_SWD  // dont use when debugging
    LPC_IOCON->SWDIO_PIO1_3  = 0xC0 | (0<<3) | 1; // LED_ROW4: PIO1_3 / SWDIO
    LPC_IOCON->SWCLK_PIO0_10 = 0xC0 | (0<<3) | 1; // LED_ROW5: PIO0_10 / SWCLK
#endif
}


static inline LPC_TMR_TypeDef * KeyboardTimer(const int _row)
{
    return (_row >= 3) ? LPC_TMR16B1 : LPC_TMR16B0;
}

static inline int8_t KeyboardTimerChannel(const int _row)
{
    return (_row >= 3) ? (_row - 3) : _row;
}

static inline int handleKeyboardColumn(const int collumn, const int rowsValue)
{
    const unsigned long mask = ~((unsigned long)7 << (collumn*KeyboardRowCount));
    const unsigned long value = rowsValue << (collumn*KeyboardRowCount);
    const unsigned long newKeyState = (currentKeyState & mask) | value;
    if (newKeyState != currentKeyState)
    {
	currentKeyState = newKeyState;
	return 1;
    }
    return 0;
}

static inline void Keyboard_SelectColumn(const int _col)
{
    int i;
    for (i = 0 ; i < KeyboardColumnCount; ++i)
    {
	const char pinLevel = (_col==i) ? ColumnSelectedLevels[i] : ColumnDeselectedLevels[i];
	if (pinLevel)
	{
	    palSetPad(columnPorts[i] ? GPIO1 : GPIO0, columnPins[i]);
	}
	else
	{
	    palClearPad(columnPorts[i] ? GPIO1 : GPIO0, columnPins[i]);
	}
    }
}


static inline void Keyboard_SetLedRows(const volatile unsigned short * _rowValues, const int _count)
{
    (void)_count; // TODO
    int i;
    for (i = 0; i < _count; ++i)
    {
	if (_rowValues[i] >= 10)
	{
	    palSetPad(ledRowPorts[i] ? GPIO1 : GPIO0, ledRowPins[i]);
	}
	int timerValue = _rowValues[i];

	if ((_rowValues[i] < 10) || (_rowValues[i] >= 990))
	{
	    timerValue = KeyboardLed_TimerInactiveValue;
	}
	Chip_TIMER_SetMatch(KeyboardTimer(i), KeyboardTimerChannel(i), timerValue);
    }
}

inline void _Keyboard_ClearLedRow(const int _row)
{
    palClearPad(ledRowPorts[_row] ? GPIO1 : GPIO0, ledRowPins[_row]);
    Chip_TIMER_SetMatch(KeyboardTimer(_row), KeyboardTimerChannel(_row), KeyboardLed_TimerInactiveValue);
}

inline void Keyboard_ClearLedRows(void)
{
    int row;
    for (row = 0; row < KeyboardLedRowCount; ++row)
    {
	_Keyboard_ClearLedRow(row);
    }
}

inline void Keyboard_ClearLedRow(const int _row)
{
    if (_row < KeyboardLedRowCount)
    {
	_Keyboard_ClearLedRow(_row);
    }
}

inline int Keyboard_ReadRows(void)
{
    int value = 0;
    int i;
    for (i = 0; i < KeyboardRowCount; ++i)
    {
	if(palReadPad(kbdRowPorts[i] ? GPIO1 : GPIO0, kbdRowPins[i]) == 0)
	    value |= 1 << i;
    }
    return value;
}

static void InitializeLedValues(void)
{
    int i;
    for (i = 0; i < KeyboardLedStateCount; ++i)
    {
	int j;
	for (j = 0; j < KeyboardColumnCount; ++j)
	{
	    int k;
	    for (k = 0; k < KeyboardLedRowCount; ++k)
	    {
		ledValues[i][j][k] = 0;
	    }
	}
    }
}

#ifdef DEBUG_AQUIRE_CHANNEL
static unsigned short daq_value[32];
static uint32_t daq_ptr = 0;
#else
static uint16_t adc_values[FaderCount][MedianFilterSize];
static uint8_t  adc_ptr;
static float    faderValue[FaderCount];
static uint16_t reportedFaderValue[FaderCount];
#endif

/**
 * @brief   CT16B0 interrupt handler.
 * @isr
 */
CH_IRQ_HANDLER(Vector80) {

  CH_IRQ_PROLOGUE();

  if (Chip_TIMER_MatchPending(LPC_TMR16B0, 3))
  {
      Chip_TIMER_ClearMatch(LPC_TMR16B0, 3);
      /*
       * Timer 16B0 has allready been stopped.
       * Need to stop 16B1 explicitly.
       */
      Chip_TIMER_Disable(LPC_TMR16B1);


      if (handleKeyboardColumn (currentKeyCollumn, Keyboard_ReadRows()))
      {
	  // act upon changes.
      }
      currentKeyCollumn = (currentKeyCollumn + 1) % KeyboardColumnCount;

      Keyboard_ClearLedRows();
      Keyboard_SelectColumn(currentKeyCollumn);
      // select the current state of the blink pattern.

      ++currentBlinkTimeout;
      if (currentBlinkTimeout >= 250)
      {
	  currentBlinkTimeout = 0;
	  currentBlinkState = (currentBlinkState + 1) % KeyboardLedStateCount;
      }
      Keyboard_SetLedRows(&ledValues[currentBlinkState][currentKeyCollumn][0], KeyboardLedRowCount);

      Chip_TIMER_Reset(LPC_TMR16B0);
      Chip_TIMER_Enable(LPC_TMR16B0);
      Chip_TIMER_Reset(LPC_TMR16B1);
      Chip_TIMER_Enable(LPC_TMR16B1);

#ifdef DEBUG_AQUIRE_CHANNEL
      {
	  unsigned short value = 0;
	  if (FaderGetAdcValue(0, &value))
	  {
	      daq_value[daq_ptr] = value;
	      ++daq_ptr;
	      if (daq_ptr >= 32)
	      {
		  daq_ptr = 0;

		  int i;
		  for (i = 0; i < 32; ++i)
		  {
		      ((unsigned short *)eventFrame.data)[i] = daq_value[i];
		  }
		  lpc1114_ump_send_frame(&eventFrame);
	      }
	  }
      }
#else
      {
	  int channel;
	  for (channel = 0; channel < FaderCount; ++channel)
	  {
	      unsigned short value = 0;
	      if (FaderGetAdcValue(channel, &value))
	      {
		  adc_values[channel][adc_ptr] = value;
	      }

	      /* spread filtering */
	      if (channel == adc_ptr)
	      {
		  const int   filteredValue = median(adc_values[channel], MedianFilterSize);
		  const float faderNew = (faderValue[channel] + faderValue[channel] + filteredValue) / 3.0;
		  faderValue[channel] = faderNew;
		  if (reportedFaderValue[channel] != (uint16_t)faderNew)
		  {
		      reportedFaderValue[channel] = (uint16_t)faderNew;
		      // report reportedFaderValue[channel].
		      if (g_inputEventMailbox)
		      {
			  const msg_t m = FADER_MESSAGE(channel, reportedFaderValue[channel]);
			  chMBPostI (g_inputEventMailbox, m);
		      }
		  }
	      }
	  }
	  adc_ptr = (adc_ptr + 1) % MedianFilterSize;
      }
#endif

  }
  else
  {
      if (Chip_TIMER_MatchPending(LPC_TMR16B0, 0))
      {
	  Chip_TIMER_ClearMatch(LPC_TMR16B0, 0);
	  Keyboard_ClearLedRow(0);
      }
      
      if (Chip_TIMER_MatchPending(LPC_TMR16B0, 1))
      {
	  Chip_TIMER_ClearMatch(LPC_TMR16B0, 1);
	  Keyboard_ClearLedRow(1);
      }
      
      if (Chip_TIMER_MatchPending(LPC_TMR16B0, 2))
      {
	  Chip_TIMER_ClearMatch(LPC_TMR16B0, 2);
	  Keyboard_ClearLedRow(2);
      }
  }
  CH_IRQ_EPILOGUE();
}

/**
 * @brief   CT16B1 interrupt handler.
 * @isr
 */
CH_IRQ_HANDLER(Vector84) {

  CH_IRQ_PROLOGUE();

  if (Chip_TIMER_MatchPending(LPC_TMR16B1, 0))
  {
      Chip_TIMER_ClearMatch(LPC_TMR16B1, 0);
      Keyboard_ClearLedRow(3);
  }

  if (Chip_TIMER_MatchPending(LPC_TMR16B1, 1))
  {
      Chip_TIMER_ClearMatch(LPC_TMR16B1, 1);
      Keyboard_ClearLedRow(4);
  }

  CH_IRQ_EPILOGUE();
}


void InitializeLedTimers(void)
{
    { // Chip_TIMER_Init(LPC_TIMER16_0);
	LPC_SYSCON->SYSAHBCLKCTRL |= (1 << 7); // Enable TIMER16_0 clock
    }

    Chip_TIMER_Reset(LPC_TMR16B0); // Chip_TIMER_Reset(LPC_TIMER16_0);

    LPC_TMR16B0->MCR = 0;
    LPC_TMR16B0->MCR |= 1<<0;  // Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 0);
    LPC_TMR16B0->MCR |= 1<<3; // Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 1);
    LPC_TMR16B0->MCR |= 1<<6;  // Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 2);
    LPC_TMR16B0->MCR |= 1<<9;  // Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 3);
    LPC_TMR16B0->MCR |= 1<<10; // Chip_TIMER_ResetOnMatchEnable(LPC_TIMER16_0, 3);
    LPC_TMR16B0->MCR |= 1<<11; // Chip_TIMER_StopOnMatchEnable(LPC_TIMER16_0, 3);

    LPC_TMR16B0->PR = 72-1; // Chip_TIMER_PrescaleSet(LPC_TIMER16_0, 72-1); //prescaler:72 = time base 1µs

    Chip_TIMER_SetMatch(LPC_TMR16B0, 3, KeyboardLed_TimerResolution); // length in µs (MAT3 )
    Chip_TIMER_SetMatch(LPC_TMR16B0, 0, KeyboardLed_TimerInactiveValue);  //low in µs
    Chip_TIMER_SetMatch(LPC_TMR16B0, 1, KeyboardLed_TimerInactiveValue);  //low in µs
    Chip_TIMER_SetMatch(LPC_TMR16B0, 2, KeyboardLed_TimerInactiveValue);  //low in µs
    Chip_TIMER_PWMWrite(LPC_TMR16B0, 0); // disable all PWM

    //-- use TIMER_16_1 for more LED's
    { // Chip_TIMER_Init(LPC_TIMER16_1);
	LPC_SYSCON->SYSAHBCLKCTRL |= (1 << 8); // Enable TIMER16_1 clock
    }

    Chip_TIMER_Reset(LPC_TMR16B1); // Chip_TIMER_Reset(LPC_TIMER16_1);

    LPC_TMR16B1->MCR = 0;
    LPC_TMR16B1->MCR |= 1<<0; // Chip_TIMER_MatchEnableInt(LPC_TIMER16_1, 0);
    LPC_TMR16B1->MCR |= 1<<3; // Chip_TIMER_MatchEnableInt(LPC_TIMER16_1, 1);

    LPC_TMR16B1->PR = 72-1; // Chip_TIMER_PrescaleSet(LPC_TIMER16_1, 72-1); //prescaler:72 = time base 1µs

    Chip_TIMER_SetMatch(LPC_TMR16B1, 0, KeyboardLed_TimerInactiveValue);  //low in µs
    Chip_TIMER_SetMatch(LPC_TMR16B1, 1, KeyboardLed_TimerInactiveValue);  //low in µs
    Chip_TIMER_PWMWrite(LPC_TMR16B1, 0); // disable all PWM

    //-- Start both timers
    Chip_TIMER_Enable(LPC_TMR16B0);
    // NVIC->ICPR[0] = (1 << ((uint32_t)(TIMER_16_0_IRQn) & 0x1F)); /* Clear pending interrupt */
    nvicEnableVector(TIMER_16_0_IRQn, CORTEX_PRIORITY_MASK(2));

    Chip_TIMER_Enable(LPC_TMR16B1);
    // NVIC->ICPR[0] = (1 << ((uint32_t)(TIMER_16_1_IRQn) & 0x1F)); /* Clear pending interrupt */
    nvicEnableVector(TIMER_16_1_IRQn, CORTEX_PRIORITY_MASK(3));
}


static WORKING_AREA(ConsoleInputThread, 180);
static msg_t ConsoleInputThreadFunction(void *arg)
{
    Mailbox *inputEventMailbox = (Mailbox *)arg;

    (void)arg;
    chRegSetThreadName("conin");

    unsigned long  oldkeystate = 0L;

    int heardbeat = 0;
    while (1)
    {
	//-- limit changes to less than 33Hz
	chThdSleepMilliseconds(30);

	//-- heardbeat
	if (++heardbeat > 33)
	{
	    heardbeat = 0;
	    const msg_t m = (14<<24); // heardbeat from console
	    chMBPost (inputEventMailbox, m, TIME_INFINITE);
	}

	//-- check for button changes.
	const uint32_t newKeystate = currentKeyState;
	if (oldkeystate != newKeystate)
	{
	    unsigned char column;
	    for (column = 0; column < KeyboardColumnCount; ++column)
	    {
		const unsigned long change = oldkeystate ^ newKeystate;
		int row;
		for (row = 0; row < KeyboardRowCount; ++row)
		{
		    const int index = row + column*KeyboardRowCount;
		    if ((change >> index) & 1)
		    {
			const int pressed = (newKeystate >> index) & 1;
			const msg_t m = FADERKEY_MESSAGE(column, row, pressed);
			chMBPost (inputEventMailbox, m, TIME_INFINITE);
		    }
		}
	    }
	    oldkeystate = newKeystate;
	}
    }
    return 0;
}

void initConsoleInput(Mailbox *inputEventMailbox)
{
    g_inputEventMailbox = inputEventMailbox;

    eventFrame.source = lpc1114_ump_my_id();
    eventFrame.dest = 1;
    eventFrame.flags = 1;
    eventFrame.size = 64;

    InitializeAdc();
    InitializeKeyboardPinmux();

    InitializeLedValues();
    InitializeLedTimers();

    chThdCreateStatic(ConsoleInputThread, sizeof(ConsoleInputThread), NORMALPRIO, ConsoleInputThreadFunction, inputEventMailbox);
}

void setLedValues(const uint8_t _fader, const uint8_t _ledno, const uint8_t _mask, const uint16_t _value)
{
    if (_fader < KeyboardColumnCount)
    {
	if (_ledno < KeyboardLedRowCount)
	{
	    int i;
	    for (i = 0; i < KeyboardLedStateCount; ++i)
	    {
		if ( (_mask & (1<<i)) != 0)
		{
		    ledValues[i][_fader][_ledno] = _value;
		}
	    }
	}
    }
}
