#include "ch.h"
#include "hal.h"
#include "ump_frame.h"
#include "ump.h"
#include "lpc1114_ump.h"
#include "lpc1114_registers.h"


static struct UmpInterface ump0;
static void serve_uart0_interrupt(const uint8_t IIRValue)
{
    if (IIRValue == IIR_RLS) // Receive Line Status
    {
        const uint8_t LSRValue = LPC_UART->LSR; // Read LSR clears the interrupt
	if (LSRValue & (LSR_BI|LSR_OE|LSR_PE|LSR_FE|LSR_RXFE))
        {
	    // broken char -> collision
	    ump_handle_rxerror(&ump0);
            LPC_UART->RBR;
        }
#if 0
        if (LSRValue & LSR_RDR)           /* Receive Data Ready */                      
        {
	    // read RBR will clear the interrupt
	    ump_handle_received_data(&ump0, LPC_UART->RBR);
        }
#endif
	if (LSRValue & LSR_THRE)
        {
	    ump_handle_transmitter_empty(&ump0);
        }
        if (LSRValue & LSR_TEMT)
        {
	    // ump_handle_transmitbuffer_empty(&ump0);
	}
    }
    else if (IIRValue & (IIR_RDA | IIR_CTI)) /* if (IIRValue & (IIR_RDA | IIR_CTI)) - Receive Data Available or Character timeout indicator */
    {
        //const uint8_t LSRValue = LPC_UART->LSR;
        //if (LSRValue & LSR_RDR)
	while (LPC_UART->LSR & LSR_RDR)
        {
	    const uint8_t c = LPC_UART->RBR;
	    ump_handle_received_data(&ump0, c);
        }
    }
    else if (IIRValue == IIR_THRE)        /* transmit holding register empty */
    {
	ump_handle_transmitter_empty(&ump0);
#if 0
        const uint8_t LSRValue = LPC_UART->LSR; // Read LSR clears the interrupt
        if (LSRValue & LSR_TEMT)
        {
	    // ump_handle_received_data(_ump, _ump->tx.expected_txdata);
	    ump_handle_transmitbuffer_empty(&ump0);
	}
#endif
    }
}

/**
 * @brief   UART0 IRQ handle.
 * @isr
 */
CH_IRQ_HANDLER(Vector94)
{
  CH_IRQ_PROLOGUE();

  uint8_t IIRValue = LPC_UART->IIR & 0x0F;
  while ((IIRValue&1) == 0)
  {
      serve_uart0_interrupt(IIRValue);

      IIRValue = LPC_UART->IIR & 0x0F;
  }

  CH_IRQ_EPILOGUE();
}

static inline void InitializeUart(void)
{
    // Enable IOCON block
    LPC_SYSCON->SYSAHBCLKCTRL |= SYSAHBCLKCTRL_IOCON;

    LPC_IOCON->PIO1_6 = 0xC1; /* RDX without resistors. */
    LPC_IOCON->PIO1_7 = 0xC1; // 0xC1 /* TDX without resistors. */

    // Enable power to UART block
    LPC_SYSCON->SYSAHBCLKCTRL |= SYSAHBCLKCTRL_UART;
    LPC_SYSCON->UARTCLKDIV = 0x01; // UARTCLK = SYSCLK / 1

    // DL = (SystemCoreClock * LPC_SYSCON->SYSAHBCLKDIV) / (16 * baudrate * LPC_SYSCON->UARTCLKDIV);

    /* Init the uart peripheral block */
    LPC_UART->LCR = 0x80; /* Enable access to Divisor Latches (DLAB=1) */
    LPC_UART->DLM = 0x00;
    LPC_UART->DLL = 0x11;
    LPC_UART->FDR = 0xF8;
    LPC_UART->LCR = 0x03; /* 8 databits, 2 stopbits, no parity (DLAB=0) */
    LPC_UART->FCR = 0x07 /* | (2<<6)*/; /* Enable and reset TX and RX FIFO, trigger level 0 */

#define UART_IRQ_PRIORITY    2
    nvicEnableVector(UART_IRQn, CORTEX_PRIORITY_MASK(LPC11xx_SERIAL_UART0_IRQ_PRIORITY));

    // LPC_UART->IER = IER_RXL | IER_RBR; /* Enable UART RX and line status interrupt */
}


static void ump_lpc1114_stop_uart (struct UmpInterface *_ump)
{
    (void)_ump;
    gptStop(&GPTD4);
    LPC_UART->IER = 0;
    nvicDisableVector(UART_IRQn);
}

static void ump_lpc1114_start_uart (struct UmpInterface *_ump)
{
    (void)_ump;
#define UART_IRQ_PRIORITY    2
//    nvicEnableVector(UART_IRQn, CORTEX_PRIORITY_MASK(LPC11xx_SERIAL_UART0_IRQ_PRIORITY));

    LPC_UART->IER = IER_RXL | IER_RBR;
}

static void ump_lpc1114_uart_send_octet(struct UmpInterface *_ump, const uint8_t _octet)
{
    (void)_ump;
    LPC_UART->THR = _octet;
}

static void _ump_timer_callback_handler(GPTDriver *gptp)
{
    // ump_lpc1114_uart_send_octet(&ump0, 'T');
    ump_handle_timer_interrupt(&ump0);
    (void)gptp;
}

void ump_lpc1114_init_timer (struct UmpInterface *_ump)
{
    (void)_ump;
    //-- configure break timer.
    static const GPTConfig sleepTimerConfig =
    {
	240000, // 240KHz Timer base = ca. 4us
	_ump_timer_callback_handler
    };
    gptStart(&GPTD4, &sleepTimerConfig);

    // ump_lpc1114_uart_send_octet(_ump, 'I');
}

static void ump_lpc1114_restart_timeout(struct UmpInterface * _ump, const uint16_t _timeout_us)
{
    (void)_ump;
    gptStartOneShotI(&GPTD4, _timeout_us / 4);
    // ump_lpc1114_uart_send_octet(_ump, 'r');
}

static void ump_lpc1114_stop_timeout(struct UmpInterface * _ump)
{
    (void)_ump;
    gptStopTimerI(&GPTD4);
    // ump_lpc1114_uart_send_octet(_ump, 's');
}


static void ump_lpc1114_frame_received(struct UmpInterface *_ump, struct UmpFrame *_frame)
{
    if ((_frame->dest == 0xff) || (_frame->dest == _ump->my_id))
    {
	_ump->frameReceivedCallback(_frame);
    }
}


//******************************************
static struct UmpCallbacks ump0_callbacks =
{
    .start_uart = ump_lpc1114_start_uart,
    .stop_uart  = ump_lpc1114_stop_uart,
    .uart_send_octet = ump_lpc1114_uart_send_octet,

    .init_timer      = ump_lpc1114_init_timer,
    .restart_timeout = ump_lpc1114_restart_timeout,
    .stop_timeout    = ump_lpc1114_stop_timeout,
    .frame_received  = ump_lpc1114_frame_received
};

//*****************************************

void lpc1114_ump_startup(void (*_frameReceivedCallback)(struct UmpFrame *))
{
    ump_init(&ump0, &ump0_callbacks, _frameReceivedCallback);
    InitializeUart();
    ump_start(&ump0);
}

void lpc1114_ump_shutdown(void)
{
    ump_stop(&ump0);
}

void lpc1114_ump_set_my_id(const uint8_t _my_id)
{
    ump_set_my_id(&ump0, _my_id);
}

uint8_t lpc1114_ump_my_id(void)
{
    return ump0.my_id;
}

void lpc1114_ump_send_frame(const struct UmpFrame *_frame)
{
    ump_send_frame(&ump0, _frame);
}

int lpc1114_ump_transmitter_ready(void)
{
    return (ump0.tx.txFrame == 0);
}
