#include "ump.h"
#include "crc8_maxim.h"

enum States
{
    Idle,
    Receive,
    ReceiveStuff,
    Transmit,
    WaitLineEmpty
};


static inline void restart_timeout(struct UmpInterface * _ump)
{
    // _ump->callbacks->restart_timeout(_ump, 13300 + _ump->my_id*100); // ID-0x11=15ms, ID-0x12=15.1ms, ID-0x13=15.2ms, ...
    _ump->callbacks->restart_timeout(_ump, 3500 + _ump->my_id*100); // ID-0x11=15ms, ID-0x12=15.1ms, ID-0x13=15.2ms, ...
}

static inline void stop_timeout(struct UmpInterface * _ump)
{
    _ump->callbacks->stop_timeout(_ump);
}


static void ump_send_next_octet(struct UmpInterface * _ump)
{
    if (_ump->tx.index >= _ump->tx.size)
    {
	// may disable THRE Interrupt here

	// change to WaitLineEmpty state:
	_ump->tx.txFrame = 0;
	restart_timeout(_ump);
	_ump->state = WaitLineEmpty;
	//++_ump->stats.txframes;
    }
    else
    {
	const uint8_t t = _ump->tx.data[_ump->tx.index];
	++_ump->tx.index;
	_ump->tx.expected_txdata = t;
	restart_timeout(_ump);
	_ump->callbacks->uart_send_octet(_ump, t);
    }
}



/*-- interrupt handlers to call --*/

void ump_handle_rxerror(struct UmpInterface *_ump)
{
    switch (_ump->state)
    {
    case Receive:
    case ReceiveStuff:
	_ump->state = WaitLineEmpty;
	restart_timeout(_ump);
	// Error during Receive
	break;

    case Transmit:
	_ump->state = WaitLineEmpty;
	restart_timeout(_ump);
	// Error during Transmit
	break;

    case WaitLineEmpty:
	// Collision on the line. Wait some more time.
	restart_timeout(_ump);
	break;

    case Idle:
	_ump->state = WaitLineEmpty;
	restart_timeout(_ump);
	break;
    }
}

static void ump_handle_startofframe(struct UmpInterface *_ump)
{
    //++_ump->stats.rx0;
    _ump->rx.rxstate = 0;
    _ump->rx.index = 0;
    _ump->rx.frame.data[0] = 'Z';
    restart_timeout(_ump);
}

static void ump_handle_rxoctet(struct UmpInterface *_ump, const uint8_t _rxdata)
{
    switch (_ump->rx.rxstate)
    {
    case 0: // expect 0xAA
	if (_rxdata != 0xAA)
	{
	    _ump->state = WaitLineEmpty;
	    restart_timeout(_ump);
	}
	else
	{
	    // ++_ump->stats.rx1;
	    _ump->rx.rxstate = 1;
	}
	break;
    case 1: // expect sender
	_ump->rx.frame.source = _rxdata;
	_ump->rx.rxstate = 2;
	break;
    case 2: // expect receiver
	_ump->rx.frame.dest = _rxdata;
	_ump->rx.rxstate = 3;
	break;
    case 3: // expect flag
	_ump->rx.frame.flags = _rxdata;
	_ump->rx.rxstate = 4;
	break;
    case 4: // expect size
	_ump->rx.frame.size = _rxdata;
	_ump->rx.rxstate = 5;
	break;
    case 5: // expect header-crc
        {
	    uint8_t headerCrc8 = 0;
	    crc8maxim_update (&headerCrc8, _ump->rx.frame.source);
	    crc8maxim_update (&headerCrc8, _ump->rx.frame.dest);
	    crc8maxim_update (&headerCrc8, _ump->rx.frame.flags);
	    crc8maxim_update (&headerCrc8, _ump->rx.frame.size);
	    if (headerCrc8 != _rxdata)
	    {
		_ump->state = WaitLineEmpty;
		restart_timeout(_ump);
		//++_ump->stats.rx3;
	    }
	    else
	    {
		if (_ump->rx.frame.size == 0)
		{
		    _ump->state = WaitLineEmpty;
		    restart_timeout(_ump);
		    //++_ump->stats.rx2;
		    _ump->callbacks->frame_received(_ump, &_ump->rx.frame);
		}
		else
		{
		    _ump->rx.crc8 = 0;
		    _ump->rx.index = 0;
		    _ump->rx.rxstate = 6;
		}
	    }
	}
	break;

    case 6: // expect body
	if (_ump->rx.index >= _ump->rx.frame.size)
	{
	    if (_ump->rx.crc8 == _rxdata)
	    {
		// Wait until line empty.
		_ump->state = WaitLineEmpty;
		restart_timeout(_ump);
		//++_ump->stats.rx2;
		_ump->callbacks->frame_received(_ump, &_ump->rx.frame);
	    }
	    else
	    {
		// Error
		//++_ump->stats.rx3;
		_ump->state = WaitLineEmpty;
		restart_timeout(_ump);
	    }
	}
	else
	{
	    crc8maxim_update (&_ump->rx.crc8, _rxdata);
	    _ump->rx.frame.data[_ump->rx.index] = _rxdata;
	    ++_ump->rx.index;
	}
	break;

    default:
	_ump->state = WaitLineEmpty;
	restart_timeout(_ump);
	break;
    }
}


void ump_handle_received_data(struct UmpInterface *_ump, const uint8_t _rxdata)
{
    //++_ump->stats.rx4;
    /* (void)_ump; uint8_t c = _rxdata; if (c >= 'a' && c <= 'c') c = c + 'A' - 'a'; _ump->callbacks->uart_send_octet(_ump, c); */

    switch (_ump->state)
    {
    case Receive:
	// check received data
	switch (_rxdata)
	{
	case 0x55:
	    //   new rx frame, did we miss something. Change to WaitLineEmpty or to Receive with new frame
	    _ump->state = Receive;
	    ump_handle_startofframe(_ump);
	    break;

	case 27:
	    _ump->state = ReceiveStuff;
	    break;

	default:
	    ump_handle_rxoctet(_ump, _rxdata);
	    break;
	}
	break;

    case ReceiveStuff:
	switch (_rxdata)
	{
	case 0x55:
	    //   new rx frame, did we miss something. Change to WaitLineEmpty or to Receive with new frame
	    _ump->state = Receive;
	    ump_handle_startofframe(_ump);
	    break;

	case 0x05:
	    _ump->state = Receive;
	    ump_handle_rxoctet(_ump, 0x55);
	    break;

	case 0x02:
	    _ump->state = Receive;
	    ump_handle_rxoctet(_ump, 27);
	    break;

	default:
	    // Error - recover from error
	    _ump->state = WaitLineEmpty;
	    restart_timeout(_ump);
	    break;
	}
	break;

    case Transmit:
	// check against expected data
	if (_rxdata == _ump->tx.expected_txdata)
	{
	    ump_send_next_octet(_ump);
	    restart_timeout(_ump);
	}
	else
	{
	    //-- initiate retransmit
	    _ump->tx.index = 0;

	    // increase txerror-counter
	    _ump->state = WaitLineEmpty;
	    restart_timeout(_ump);
	}
	break;

    case WaitLineEmpty:
    case Idle:
	if (_rxdata == 0x55)
	{
	    _ump->state = Receive;
	    ump_handle_startofframe(_ump);
	}
	else
	{
	    _ump->state = WaitLineEmpty;
	    restart_timeout(_ump);
	}
	break;
    }
}

void ump_handle_transmitter_empty(struct UmpInterface *_ump)
{
    switch (_ump->state)
    {
    case Receive:
    case ReceiveStuff:
	// do nothing
	break;
    case Transmit:
//ump_send_next_octet(_ump); // for debugging only
	// do nothing, as we expect to receive the data
	break;
    case WaitLineEmpty:
	// do nothing
	break;
    case Idle:
	// do nothing
	break;
    }
}


void ump_handle_transmitbuffer_empty(struct UmpInterface *_ump)
{
    (void)_ump;
    // do nothing, as we can not decide what to do next
}



void ump_handle_timer_interrupt(struct UmpInterface *_ump)
{
    switch (_ump->state)
    {
    case Receive:
    case ReceiveStuff:
	//++_ump->stats.to0;
	//_ump->stats.to3 = _ump->rx.index;
	_ump->state = WaitLineEmpty;
	restart_timeout(_ump);
	// Timeout during Receive -> incomplete frame -> WaitLineEmpty? | Idle?
	break;
    case Transmit:
	//++_ump->stats.to1;
	_ump->state = WaitLineEmpty;
	_ump->tx.txFrame = 0; // Abort
	restart_timeout(_ump);
	// Timeout during Transmit -> no echo -> line error
	break;
    case WaitLineEmpty:
	//++_ump->stats.to2;
	_ump->state = Idle;
	stop_timeout(_ump);
	if (_ump->tx.txFrame != 0)
	{
	    _ump->state = Transmit;
	    // may enable THRE Interrupt here
	    ump_send_next_octet(_ump);
	}
	// Timeout during WaitLineEmpty -> line is empty
	break;
    case Idle:
	stop_timeout(_ump);
	// Timeout during Idle -> Idle
	break;
    }
}


static void ump_start_transmision(struct UmpInterface *_ump)
{
    if (_ump->state == Idle)
    {
	_ump->state = Transmit;
	// may enable THRE Interrupt here
	ump_send_next_octet(_ump);
    }
}


void ump_init(struct UmpInterface *_ump, const struct UmpCallbacks *_callbacks, void (*_frameReceivedCallback)(struct UmpFrame *))
{
    _ump->state = 0;
    _ump->callbacks = _callbacks;
    _ump->frameReceivedCallback = _frameReceivedCallback;
    _ump->tx.index = 0;
    _ump->tx.size = 0;
#if 0
    _ump->stats.txframes = 0;
    _ump->stats.rx0 = 0;
    _ump->stats.rx1 = 0;
    _ump->stats.rx2 = 0;
    _ump->stats.rx3 = 0;
    _ump->stats.rx4 = 0;
#endif
}

void ump_set_my_id(struct UmpInterface *_ump, const uint8_t _my_id)
{
    _ump->my_id = _my_id;
}


void ump_start(struct UmpInterface *_ump)
{
    _ump->callbacks->init_timer(_ump);
    _ump->callbacks->start_uart(_ump);
}

void ump_stop(struct UmpInterface *_ump)
{
    _ump->callbacks->stop_uart(_ump);
}

static void tx_add_stuffed_byte(struct UmpInterface *_ump, const uint8_t _src)
{
    if (_src == 27)
    {
	_ump->tx.data[_ump->tx.size] = 27;
	++_ump->tx.size;
	_ump->tx.data[_ump->tx.size] = 0x02;
	++_ump->tx.size;
    }
    else if (_src == 0x55)
    {
	_ump->tx.data[_ump->tx.size] = 27;
	++_ump->tx.size;
	_ump->tx.data[_ump->tx.size] = 0x05;
	++_ump->tx.size;
    }
    else
    {
	_ump->tx.data[_ump->tx.size] = _src;
	++_ump->tx.size;
    }
}

void ump_send_frame(struct UmpInterface *_ump, const struct UmpFrame * _txFrame)
{
    if (_ump->tx.txFrame == 0)
    {
	if (_txFrame->size > 0)
	{
	    _ump->tx.data[0] = 0x55;
	    _ump->tx.data[1] = 0xAA;
	    _ump->tx.size = 2;

	    tx_add_stuffed_byte(_ump, _txFrame->source); // source
	    tx_add_stuffed_byte(_ump, _txFrame->dest); // destination
	    tx_add_stuffed_byte(_ump, _txFrame->flags); // flags
	    tx_add_stuffed_byte(_ump, _txFrame->size);

	    {
		uint8_t headerCrc8 = 0;
		crc8maxim_update (&headerCrc8, _txFrame->source);
		crc8maxim_update (&headerCrc8, _txFrame->dest);
		crc8maxim_update (&headerCrc8, _txFrame->flags);
		crc8maxim_update (&headerCrc8, _txFrame->size);
		tx_add_stuffed_byte(_ump, headerCrc8);
	    }

	    {
		uint8_t payloadCrc8 = 0;
		int i;
		for (i = 0; i < _txFrame->size; ++i)
		{
		    crc8maxim_update (&payloadCrc8, _txFrame->data[i]);
		    tx_add_stuffed_byte(_ump, _txFrame->data[i]);
		}
		tx_add_stuffed_byte(_ump, payloadCrc8);
	    }

	    _ump->tx.txFrame = _txFrame;
	    _ump->tx.index = 0;

	    ump_start_transmision(_ump);
	}
    }
}
